from decimal import Decimal

INT_MAX = 999999
INT_MIN = -999999
REAL_MAX = Decimal("999999.9999") 
REAL_MIN = Decimal("-999999.9999")
STRING_MAX_LENGTH = 300
LABEL_MAX_LENGTH = 40
HELP_TEXT_MAX_LENGTH = 200

